/*
 GBxCart RW - 24 Switch Puzzle Cart
 Version: 1.1
 Author: Alex from insideGadgets (www.insidegadgets.com)
 Created: 18/10/2021
 Last Modified: 3/04/2022
 License: CC-BY-NC-SA
 
 Re-program the switch combination on the 24 Switch Puzzle cart
 
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <time.h>
#include <signal.h>

#ifdef _WIN32
#include <windows.h>
#else
#include <unistd.h>
#endif

#include "setup.h" // See defines, variables, constants, functions here

void sigintHandler(int sig_num) {
	set_mode('0');
	RS232_flushRX(cport_nr);
	gbx_cart_power_down();
	exit(0);
}

int main(int argc, char **argv) {
	signal(SIGINT, sigintHandler);
	
	printf("24-Switch Puzzle Cart v1.1 by insideGadgets\n");
	printf("###########################################\n");
	
	read_config();
	
	// Open COM port
	if (com_test_port() == 0) {
		printf("Device not connected and couldn't be auto detected. Please make sure the COM port isn't open elsewhere. If you are using Linux/Mac, make sure you use \"sudo\".\n");
		read_one_letter();
		return 1;
	}
	printf("Connected on COM port: %i at ", cport_nr+1);
	
	// Break out of any existing functions on ATmega
	set_mode('0');
	RS232_flushRX(cport_nr);
	
	// Get cartridge mode - Gameboy or Gameboy Advance
	cartridgeMode = request_value(CART_MODE);
	
	// Get PCB version
	gbxcartPcbVersion = request_value(READ_PCB_VERSION);
	xmas_wake_up();
	
	// Increase to 1.7Mbit baud for v1.4 PCB
	#ifdef _WIN32
	if ((gbxcartPcbVersion >= PCB_1_4 && gbxcartPcbVersion < GBXMAS) || gbxcartPcbVersion == MINI_1_1) {
		set_mode(USART_1_7M_SPEED);
		RS232_CloseComport(cport_nr);
		delay_ms(200);
		
		if (RS232_OpenComport(cport_nr, 1700000, "8N1") == 0) {
			printf("1.7M Baud\n");
		}
	}
	else {
		printf("1M Baud\n");
	}
	#else
		printf("1M Baud\n");
	#endif
	
	char pcbVersionString[20];
	if (gbxcartPcbVersion == PCB_1_0) {
		strncpy(pcbVersionString, "v1.0", 5);
	}
	else if (gbxcartPcbVersion == PCB_1_1) {
		strncpy(pcbVersionString, "v1.1/v1.2", 10);
	}
	else if (gbxcartPcbVersion == PCB_1_3) {
		strncpy(pcbVersionString, "v1.3", 5);
	}
	else if ((gbxcartPcbVersion >= PCB_1_4 && gbxcartPcbVersion < GBXMAS)) {
		strncpy(pcbVersionString, "v1.4", 5);
	}
	else if (gbxcartPcbVersion == MINI) {
		strncpy(pcbVersionString, "Mini", 5);
	}
	else if (gbxcartPcbVersion == MINI_1_1) {
		strncpy(pcbVersionString, "Mini v1.1", 10);
	}
	else if (gbxcartPcbVersion == GBXMAS) {
		strncpy(pcbVersionString, "XMAS", 5);
	}
	
	// Get firmware version
	gbxcartFirmwareVersion = request_value(READ_FIRMWARE_VERSION);
	printf("PCB version: %s, Firmware version: %i\n", pcbVersionString, gbxcartFirmwareVersion);
	
	if (gbxcartFirmwareVersion == 0) {
		printf("\nFirmware L1 may be installed. In order to use this applications you will need to downgrade to R30.\n");
		read_one_letter();
		return 1;
	}
	
	set_mode(VOLTAGE_5V);
	delay_ms(50);
	
	// GBx v1.4 - Power up the cart if not already powered up and flush buffer
	gbx_cart_power_up();
	RS232_flushRX(cport_nr);
	
	// Get cartridge mode - Gameboy or Gameboy Advance
	cartridgeMode = request_value(CART_MODE);
	read_gb_header();
	
	
	
	uint8_t switch8First = 0;
	uint8_t switch8Second = 0;
	uint8_t switch8Third = 0;
	
	printf("\nYou must ensure that the switches are in the correct position to unlock the cart otherwise the below won't work.\n");
	printf("\n*** It's very important to write down the switch combination or screenshot it otherwise you won't be able to re-program the switches.***\n\n");
	printf("\nThis will re-program the switches on the cart.\nPlease enter a 0 (left position) or 1 (right position).\n\n");
	printf("                                     ########\n");
	
	printf("Enter position for first 8 switches: ");
	char switchPositionString[10];
	fgets(switchPositionString, 10, stdin);
	
	for (uint8_t x = 0; x < 8; x++) {
		if (switchPositionString[x] == '1') {
			switch8First |= (1<<x);
		}
	}
	
	printf("Enter position for second 8 switches:");
	fgets(switchPositionString, 10, stdin);
	
	for (uint8_t x = 0; x < 8; x++) {
		if (switchPositionString[x] == '1') {
			switch8Second |= (1<<x);
		}
	}
	
	printf("Enter position for third 8 switches: ");
	fgets(switchPositionString, 10, stdin);
	
	for (uint8_t x = 0; x < 8; x++) {
		if (switchPositionString[x] == '1') {
			switch8Third |= (1<<x);
		}
	}
	
	
	printf("\nThe switch combinations are:\n");
	for (uint8_t x = 0; x < 8; x++) {
		if (switch8First & (1<<x)) {
			printf("1");
		}
		else {
			printf("0");
		}
	}
	printf("\n");
	for (uint8_t x = 0; x < 8; x++) {
		if (switch8Second & (1<<x)) {
			printf("1");
		}
		else {
			printf("0");
		}
	}
	printf("\n");
	for (uint8_t x = 0; x < 8; x++) {
		if (switch8Third & (1<<x)) {
			printf("1");
		}
		else {
			printf("0");
		}
	}
	printf("\n\n");
	
	printf("Press enter twice to confirm and write to the cart\n");
	read_one_letter();
	
	
	// Unlock cart
	set_bank(0xB000, 0);
	delay_ms(50);
	
	// Set switch data
	set_bank(0xB000, switch8First);
	delay_ms(50);
	set_bank(0xB000, switch8Second);
	delay_ms(50);
	set_bank(0xB000, switch8Third);
	delay_ms(500);
	
	// GBx v1.4 - Power down cart
	gbx_set_done_led();
	gbx_cart_power_down();
	
	printf("Completed.\n");
	read_one_letter();
	
	
	return 0;
}